-- This file is part of the reachability analysis tool LaRA
-- (Lazy Reachability Analyzer)
--
-- Loïg Jezequel & Didier Lime
--
-- University of Nantes, École Centrale de Nantes, All rights reserved, 2016

module Lara.Data (
    PNode(..),
    Automaton(..),
    Network(..),
    Config(..),
    Objective,
    showAuto,
    updateLabel,
    initial,
    mergeNodes,
    gLab,
    independent
    ) where

import Data.List
import qualified Data.Map.Strict as M
import qualified Data.Set as S

import qualified Data.Graph.Inductive.Graph as G
import qualified Data.Graph.Inductive.PatriciaTree as GP

import Control.Applicative

--------------------------------------------------------------------------------
--                               Data-structures
--------------------------------------------------------------------------------
type Config = [Int]

-- a PAutomaton Node
data PNode = PNode { config::Config, goal::Bool, autoReqs::S.Set Int, label::String} deriving Eq

-- An automaton representing the (partial) product of automata or just one automaton
data Automaton = Automaton { graph::GP.Gr PNode Int, nodeMap::M.Map Config G.Node, autoIDs::[Int], goals::(S.Set G.Node), name::String , laReqs::(M.Map Int (S.Set Int)), hasObj::Bool} deriving (Eq, Show)


type Objective = [(Int,Int)]
data Network = Network { automata::[Automaton], labels::[String], objective::Objective, lnReqs::(M.Map Int (S.Set Int))}
   

--------------------------------------------------------------------------------
--                                  Display
--------------------------------------------------------------------------------
showAuto :: Network -> Automaton -> String
showAuto net aut = let g = graph aut
                       ns = G.nodes g
                       ls = labels net
                       showEdge (n',i) = let pn' = (\(Just x)->x) $ G.lab g n' in "\n    -- " ++ (ls !! i) ++ " -> " ++ show pn'
                       showNode n = let pn = (\(Just x)->x) $ G.lab g n in show pn  ++ concatMap showEdge (G.lsuc g n) ++ "\n"
                    in concatMap showNode ns

instance Show PNode where
    show pn = label pn ++ (if goal pn then "!" else "") -- ++ " " ++ (show $ S.toList $ autoReqs pn)


--------------------------------------------------------------------------------
--                                  Misc
--------------------------------------------------------------------------------

gLab :: GP.Gr PNode Int -> G.Node -> PNode
gLab g = (\(Just x)->x).(G.lab g)


-- update the label of a node in a graph
updateLabel :: Int -> PNode -> GP.Gr PNode Int -> GP.Gr PNode Int
updateLabel n l' g = case G.match n g of
      (Just (p, _, l, s), cg) -> (p, n, l', s) G.& cg
      (Nothing, _)            -> error $ "Unknown node" ++ show n


-- initial node of a Automaton
initial :: Automaton -> Config
initial = (map (const 0)).autoIDs

-- merge helper
mergeN' :: [(Int,Int)] -> [(Int,Int)] -> Maybe [(Int,Int)]
mergeN' [] ys = Just ys
mergeN' xs [] = Just xs
mergeN' xs@((i1,s1):xs') ys@((i2,s2):ys') = 
    if i1 == i2 then if s1 == s2 then liftA2 (:) (Just (i1,s1)) (mergeN' xs' ys')
                                 else Nothing
                else if i1 < i2 then liftA2 (:) (Just (i1,s1)) (mergeN' xs' ys)
                                else liftA2 (:) (Just (i2,s2)) (mergeN' xs ys')

-- Merge Config's from two Automaton's
mergeNodes :: Config -> [Int] -> Config -> [Int] -> Maybe [Int]
mergeNodes c1 as1 c2 as2 = 
    let 
        x1 = zip as1 c1
        x2 = zip as2 c2 
    in
        fmap (map snd) $ mergeN' x1 x2


-- is action i from aut1 independent from aut2?
independent :: Automaton -> Int -> Automaton -> Bool
independent aut1 i aut2 =
    let sas2 = S.fromList $ autoIDs aut2
     in S.null $ S.intersection sas2 $ M.findWithDefault S.empty i (laReqs aut1)
        

