-- This file is part of the reachability analysis tool LaRA
-- (Lazy Reachability Analyzer)
--
-- Loïg Jezequel & Didier Lime
--
-- University of Nantes, École Centrale de Nantes, All rights reserved, 2016

module Lara.PReach (preach,preach',prepare,autoProductInit,autoProduct,autoProduct1e) where

import Data.Ord
import Data.Function
import Data.List
import qualified Data.Graph.Inductive.Graph as G
import qualified Data.Graph.Inductive.PatriciaTree as GP
import qualified Data.Set as S
import qualified Data.Map.Strict as M
import qualified Data.Sequence as Q

import Control.Applicative

import System.Environment
import Lara.Parser
import Debug.Trace

import Lara.Data

import Control.Parallel.Strategies

goalTrim' :: Automaton -> [G.Node] -> Automaton -> Automaton
goalTrim' aut []    autr = autr
goalTrim' aut (n':w) autr = 
    let
        g  = graph aut
        gr = graph autr
        -- predecessors (with label of transitions) of the current node
        ps = G.lpre g n'
        -- extend to real edges by adding target state n'
        es = map (\(n,i) -> (n,n',i)) ps
        -- predecessor nodes that are not already in the graph
        ns = map fst $ filter (\(n,i) -> not $ G.gelem n gr) ps
        w' = ns ++ w
        -- extend ns with node labels
        ls = map (\n->(n, gLab g n)) ns
        autr' = autr { graph = G.insEdges es $ G.insNodes ls gr }
    in
        goalTrim' aut w' autr' 

-- Keep only the states that are coreachable for the goals
goalTrim :: Automaton -> Automaton
goalTrim aut = 
    let w = S.toList $ goals aut
        g  = graph aut
        -- extend list of goal states with node labels
        ls = map (\n->(n, gLab g n)) w
        autr = aut { graph = G.insNodes ls G.empty }
     in goalTrim' aut w autr

-- update the automata requirements for states for one edge
updateReqs1e :: G.Node -> (Automaton, S.Set G.Node, [G.Node]) -> (G.Node, Int) -> (Automaton, S.Set G.Node, [G.Node])
updateReqs1e n (aut,v,w) (n',i) = 
    let
        g   = graph aut
        sas = S.fromList $ autoIDs aut
     in 
        if S.member n' v then (aut,v,w)
                         else let 
                                  pn  = gLab g n
                                  pn' = gLab g n'
                                  reqsi = M.findWithDefault S.empty i $ laReqs aut
                               in (aut { graph = updateLabel n' pn'{ autoReqs = S.difference (S.union (autoReqs pn) reqsi) sas } g},
                                   S.insert n' v,
                                   n':w)

-- update helper
updateReqs' :: (Automaton, S.Set G.Node, [G.Node]) -> (Automaton, S.Set G.Node, [G.Node])
updateReqs' r@(aut,v,[]) = r
updateReqs' r@(aut,v,n:w) = 
    let
        g   = graph aut
        ns' = G.lsuc g n
     in 
        updateReqs' $ foldl' (updateReqs1e n) (aut,v,w) ns'


-- update the automata requirements for states
-- do not compute the whole set but only some of them (at least path to each reachable state)
updateReqs :: Automaton -> Automaton
updateReqs aut = 
    let
        g = graph aut
        (aut', _, _) = updateReqs' (aut,S.singleton 0,[0])
     in aut'
         

-- product of the two edges (n1,n1',i) of aut1 and (n2,n2',i) of aut2 starting from node n (which corresponds to (n1,n2) in in the current partial product aut). 
-- can do nothing if the two edges are not compatible
-- return the possibly completed product with possibly the edge added and wether it leads to a new node
autoProduct1e :: Automaton ->  Automaton -> G.Node -> G.Node -> G.Node -> G.Node -> Int -> G.Node -> Automaton -> (Automaton, Maybe (G.LEdge Int), Bool)
autoProduct1e aut1 aut2 n1 n1' n2 n2' i n aut =
    let 
        g1  = graph aut1
        g2  = graph aut2
        g   = graph aut
        pn1' = gLab g1 n1'
        pn2' = gLab g2 n2'
        as1 = autoIDs aut1
        as2 = autoIDs aut2
        sas = S.fromList $ autoIDs aut
        mc' = mergeNodes (config pn1') as1 (config pn2') as2
     in 
        if mc' == Nothing 
           then (aut, Nothing, False) -- these two edges do not lead to a consistent state
           else -- we have a consistent state: does it already exist?
                 let c' = (\(Just x)->x) mc'
                     maut = nodeMap aut
                     mn' = M.lookup c' maut
                  in 
                     if mn' /= Nothing 
                        then -- it exists: just add an edge
                             let n' = (\(Just x)->x) mn'
                              in (aut { graph = G.insEdge (n,n',i) g }, Just (n,n',i), False)
                        else -- it does not exist: add a new node and an edge
                             let n'   = head $ G.newNodes 1 g
                                 -- supposing we start from objectives and add to the right in the product
                                 -- so left always has an objective
                                 goal' = goal pn1' && ((not $ hasObj aut2) || goal pn2')
                                 pn = gLab g n
                                 reqsi = S.difference (M.findWithDefault S.empty i (laReqs aut)) sas

                                 reqs' = S.union (autoReqs pn) reqsi                                 
                                 pn' = PNode c' goal' reqs' (label pn1' ++ ", " ++ label pn2')
                                 aut' = aut { graph   = G.insEdge (n,n',i) $ G.insNode (n',pn') g,
                                              nodeMap = M.insert c' n' maut,
                                              goals = if goal' then S.insert n' (goals aut) 
                                                               else goals aut }
                              in (aut', Just (n,n',i), True)


autoProduct1' :: Automaton ->  Automaton -> G.Node -> G.Node -> G.Node -> (Automaton,[(G.Node,G.Node,Int)]) -> (G.Node,G.Node, Int) -> (Automaton,[(G.Node,G.Node,Int)])
autoProduct1' aut1 aut2 n1 n2 n (aut,w) (n1',n2',i) = 
    let (aut', me, b) = autoProduct1e aut1 aut2 n1 n1' n2 n2' i n aut
        w' = if me /= Nothing && b then let Just (_,n',_) = me 
                                            g = graph aut'
                                            rlength = (S.size.autoReqs.(gLab g).(\(_,_,x)->x))
                                         in  insertBy (comparing rlength) (n1',n2',n') w 
                                            --(n1',n2',n'):w
                                   else w
     in (aut', w')

autoProduct1 :: Automaton ->  Automaton -> G.Node -> G.Node -> G.Node -> (Automaton,[(G.Node,G.Node,Int)]) -> [(G.Node, G.Node, Int)] -> (Automaton,[(G.Node,G.Node,Int)])
autoProduct1 aut1 aut2 n1 n2 n = foldl' (autoProduct1' aut1 aut2 n1 n2 n) 

-- product from two nodes
-- w is the waiting list of nodes to explore in each automata
autoProductN :: Automaton ->  Automaton -> G.Node -> G.Node -> G.Node -> (Automaton,[(G.Node,G.Node,Int)]) -> (Automaton,[(G.Node,G.Node,Int)])
autoProductN aut1 aut2 n1 n2 n (aut,w) = 
    let 
        g1  = graph aut1
        g2  = graph aut2
        ss1 = G.lsuc g1 n1
        ss2 = G.lsuc g2 n2
        -- synchronised edges
        ses = [(n1',n2',i1) | (n1',i1) <- ss1, (n2',i2) <- ss2, i1 == i2] 
        -- local edges
        sas1 = S.fromList $ autoIDs aut1
        sas2 = S.fromList $ autoIDs aut2
        les1 = [(n1',n2,i) | (n1',i) <- ss1, independent aut1 i aut2] 
        les2 = [(n1,n2',i) | (n2',i) <- ss2, independent aut2 i aut1] 
        llength (_,_,x) = S.size $ S.difference (M.findWithDefault S.empty x (laReqs aut)) $ S.fromList $ autoIDs aut
     in 
        autoProduct1 aut1 aut2 n1 n2 n (aut,w) $ sortBy (comparing llength) (ses ++ les1 ++ les2)      
   
autoProduct' :: Automaton ->  Automaton -> (Automaton,[(G.Node,G.Node,Int)]) -> (Automaton,[(G.Node,G.Node,Int)])
autoProduct' aut1 aut2 (aut,[]) = (aut,[])
autoProduct' aut1 aut2 (aut,(n1,n2,n):w') = autoProduct' aut1 aut2 $ autoProductN aut1 aut2 n1 n2 n (aut,w')

autoProductInit :: Automaton -> Automaton -> Automaton
autoProductInit aut1 aut2 = 
    let
        g1  = graph aut1
        g2  = graph aut2
        pn1 = gLab g1 0
        pn2 = gLab g2 0
        as = nub $ sort $ autoIDs aut1 ++ autoIDs aut2
        c  = replicate (length as) 0
        goalI = goal pn1 && ((not $ hasObj aut2) || goal pn2)
        pn = PNode c goalI S.empty (label pn1 ++ ", " ++ label pn2)
        m  = M.singleton c 0
        g  = G.insNode (0,pn) G.empty
        goalsI = if goalI then S.singleton 0 else S.empty
        sas = S.fromList as
        -- laReqsI = M.filter (not . S.null) $ M.unionWith (\x1 x2 -> S.union x1 x2) (laReqs aut1) (laReqs aut2)
        laReqsI = M.filter (not . S.null) $ M.union (laReqs aut1) (laReqs aut2)
     in 
        Automaton g m as goalsI (name aut1 ++ ", " ++ name aut2) laReqsI True

-- product of two Automaton's
autoProduct :: Automaton -> Automaton -> Automaton
autoProduct aut1 aut2 = fst $ autoProduct' aut1 aut2 (autoProductInit aut1 aut2, [(0,0,0)])

-- Given the partial product aut, look if adding some more automata is needed (and if so do it) or if the objective cannot be fullfilled (then give Nothing)
concretise :: Network -> Automaton -> Maybe Automaton
concretise net aut = 
    let 
        g = graph aut
        (noreqs,reqs) = S.partition S.null $ S.map (autoReqs . (gLab g)) (goals aut)
        a' = head $ S.toList $ minimumBy (comparing S.size) $ S.toList $ reqs
        aut2' = (automata net !! a') 
        aut2  = aut2' { laReqs = M.mapWithKey (\i _ -> M.findWithDefault S.empty i (lnReqs net)) (laReqs aut2') }
     in 
        if S.null (goals aut) then Nothing
                              else if S.null noreqs then concretise net $ goalTrim $ autoProduct aut aut2
                                                    else Just aut

-- Do the product of overlapping automata
parProduct :: Network -> [Automaton] -> Maybe [Automaton]
parProduct _ []       = Just []
parProduct net (aut:xs) = let as = autoIDs aut 
                              indep1 = null.(intersect as).autoIDs
                              indep  = and $ map indep1 xs
                           in if indep then (fmap (aut:)) (parProduct net xs)
                                       else (sequence $ map (\x -> if indep1 x then Just x 
                                                                               else concretise net $ autoProduct aut x) xs) >>= parProduct net

-- update goals and automata requirements in automaton number a for objective state s
prepare :: Network -> (Int, Int) -> Automaton
prepare net (a,s) =
    let 
        aut = (automata net) !! a
        g   = graph aut
        pn  = gLab g s 
    in 
        updateReqs $ aut { graph = updateLabel s pn{ goal = True } g,
                           goals = S.insert s (goals aut),
                           laReqs = M.mapWithKey (\i _ -> M.findWithDefault S.empty i (lnReqs net)) (laReqs aut),
                           hasObj = True
                         }


-- Possibly give a list of independent partial products of the automata in the Network satisfying the objective and Nothing there exists no such partial product. Concretise first, then product 
preach :: Network -> Maybe [Automaton]
preach net =
    let 
        obj         = objective net
        objAut      = (map ((concretise net).(prepare net)) obj) -- `using` rpar
     in
        (sequence $ objAut) >>= parProduct net

-- Possibly give a partial product of the automata in the Network satisfying the objective and Nothing there exists no such partial product. Product first, then concretise
preach' :: Network -> Maybe [Automaton]
preach' net = 
    let 
        obj         = objective net
        objAut      = map (prepare net) obj
        initProduct = foldl1' (\x y -> goalTrim $ autoProduct x y) objAut
     in 
        sequence $ [concretise net initProduct]


