-- This file is part of the reachability analysis tool LaRA
-- (Lazy Reachability Analyzer)
--
-- Loïg Jezequel & Didier Lime
--
-- University of Nantes, École Centrale de Nantes, All rights reserved, 2016

{-# LANGUAGE FlexibleContexts #-}

module Lara.Parser (parseA) where

import qualified Text.Parsec as P
import Data.List (elemIndex,sortBy)
import Data.Ord (comparing)

import qualified Data.Foldable as F
import qualified Data.Set as S
import qualified Data.Map.Strict as M
import qualified Data.Sequence as D
import qualified Data.Graph.Inductive.Graph as G

import Lara.Data

import Debug.Trace

--------------------------------------------------------------------------------
--                                  Parser
--------------------------------------------------------------------------------

-- Ignore a space block
skipSpaces p = do x<-p; P.spaces; return x

-- Parse a character ignoring trailing spaces
symbol c = P.char c >> P.spaces

-- Parse an identifier: letter then letters or digits
identifier::P.Parsec String (M.Map String Int, D.Seq String,M.Map String Int, D.Seq String,Int,M.Map Int (S.Set Int)) String
identifier = skipSpaces $ do x<-P.letter; xs<-P.many P.alphaNum; return (x:xs)

-- Parse a number
number::(Num a, Integral a, Read a)=>P.Parsec String (M.Map String Int, D.Seq String,M.Map String Int, D.Seq String,Int,M.Map Int (S.Set Int)) a
number = skipSpaces $ fmap read $ P.many1 P.digit

-- Parse an automaton
pautomaton::P.Parsec String (M.Map String Int, D.Seq String,M.Map String Int, D.Seq String,Int,M.Map Int (S.Set Int)) Automaton
pautomaton = do
    skipSpaces (P.string "auto")
    x <- identifier
    symbol '{'
    xs <-ptransitions
    symbol '}'
    (mapstates,states,maplabs,labels,na,lu) <- P.getState
    P.modifyState (\(mx,x,my,y,n,lus)->(M.empty,D.empty,my,y,n+1,lus))

    let nodes = [0..(D.length states - 1)]
    let g = G.insEdges xs $ G.insNodes (map (\n -> (n,PNode [n] False S.empty (x ++ "." ++ D.index states n))) nodes) G.empty
    let m = M.fromList [([n],n) | n<-nodes]
    let ls = map (\(n,n',i)->(i,S.empty)) xs
    return (Automaton g m [na] S.empty x (M.fromList ls) False)

-- Parse an indentifier, add it if it does not exists and return its index in the global name list
findOrAdd::String->M.Map String Int->(Int,M.Map String Int)
findOrAdd i ns = 
    let s         = M.size ns
        f k nv ov = ov 
        (ni,ns')  = M.insertLookupWithKey f i s ns
        in if (ni == Nothing) then (s,ns')
                              else ((\(Just x)->x) ni, ns')

-- Parse a transition
ptransition::P.Parsec String (M.Map String Int, D.Seq String,M.Map String Int, D.Seq String,Int,M.Map Int (S.Set Int)) (G.LEdge Int)
ptransition = do
    (mapstates,states,maplabs,labels,n,lu) <- P.getState
    let n1 = D.length states
    let n2 = D.length labels

    -- parse a state and add its name to the list if necessary
    i <- identifier
    let (xi,ms') = findOrAdd i mapstates
    P.modifyState (\c@(mx,x,my,y,n,l)->if xi == n1 then (ms',x D.|> i,my,y,n,l) else c)

    skipSpaces (P.string "--")
    
    -- parse a label and add its name to the list if necessary
    -- also update the set of automata using this label
    l <- identifier
    let (xl,ml') = findOrAdd l maplabs
    P.modifyState (\c@(mx,x,my,y,n,lus)->if xl == n2 then (mx, x, ml', y D.|> l,n, M.insert xl (S.singleton n) lu) 
                                                     else (mx, x, my , y       ,n, M.update (\ss->Just (S.insert n ss)) xl lus))
    
    skipSpaces (P.string "->")
    
    (mapstates,states,maplabs,labels,n,lu) <- P.getState
    let n1 = D.length states

    -- parse a state and add its name to the list if necessary
    o <- identifier
    let (xo,ms'') = findOrAdd o mapstates
    P.modifyState (\c@(mx,x,my,y,n,l)->if xo == n1 then (ms'', x D.|> o, my, y, n, l) else c)

    return (xi,xo,xl)

-- Parse a list of transitions
ptransitions::P.Parsec String (M.Map String Int, D.Seq String,M.Map String Int, D.Seq String,Int,M.Map Int (S.Set Int)) [G.LEdge Int]
ptransitions = P.many1 ptransition

-- Parse a state description of the form auto.state
pautoloc::P.Parsec String (M.Map String Int, D.Seq String,M.Map String Int, D.Seq String,Int,M.Map Int (S.Set Int)) (String,String)
pautoloc = do
    a <- identifier
    symbol '.'
    s <- identifier
    return (a,s)

autoByName::[Automaton]->String->Int
autoByName as n =
    let ma = elemIndex n (map name as)
    in if ma == Nothing then error ("Automaton " ++ n ++ " unknown")
                        else (\(Just x)->x) ma
        
stateByName::Automaton->String->Int
stateByName aut s = 
    let lns = G.labNodes $ graph aut
        s' = name aut ++ "." ++ s
        match = filter ((==s').label.snd) lns
    in if null match then error ("State " ++ s ++ " unknown")
                     else fst $ head $ match
        
-- Parse an objective
pobjective::[Automaton]->P.Parsec String (M.Map String Int, D.Seq String,M.Map String Int, D.Seq String,Int,M.Map Int (S.Set Int)) [(Int,Int)]
pobjective as = do
    skipSpaces (P.string "objective")
    os <- pautoloc `P.sepBy1` (symbol '&')
    return $ sortBy (comparing fst) $ map (\(x,y)->let a = autoByName as x in (a, stateByName (as!!a) y)) os
    

-- Parse a list of automata
pautomata::P.Parsec String (M.Map String Int, D.Seq String,M.Map String Int, D.Seq String,Int,M.Map Int (S.Set Int)) Network
pautomata = do
    as <- P.many1 pautomaton
    obj <- P.option [(0,-1)] (pobjective as) -- (0,-1) corresponds to False since it can never be verified

    (mapstates,st,maplabs,labels,n,lu) <- P.getState
    let ls = [0..((D.length $ labels)-1)]
    return (Network as (F.toList labels) obj lu)


-- Main parsing function
parseA::String->Either P.ParseError Network
parseA = P.runParser (P.spaces >> pautomata) (M.empty,D.empty,M.empty,D.empty,0,M.empty) "" 

--------------------------------------------------------------------------------
--                              End of Parser
--------------------------------------------------------------------------------


